# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/23 4:10 PM
# @File: __init__.py
# @Email: mlshenkai@163.com
from .transformer import BertClassifier
from .fasttext import FastTextClassifier
from .textcnn import TextCNNClassifier
from .textrnn import TextRNNClassifier
from .ml import ClassicClassifier, TextCluster



# # -*- coding: utf-8 -*-
# # @Author: watcher
# # @Created Time: 2023/3/23 4:10 PM
# # @File: __init__.py
# # @Email: mlshenkai@163.com
# import sys
# from asyncio import AbstractEventLoop
# from datetime import datetime, time, timedelta
# from logging import Handler
# from types import TracebackType
# from typing import (
#     Any,
#     BinaryIO,
#     Callable,
#     Dict,
#     Generator,
#     Generic,
#     List,
#     NamedTuple,
#     Optional,
#     Pattern,
#     Sequence,
#     TextIO,
#     Tuple,
#     Type,
#     TypeVar,
#     Union,
#     overload,
# )
#
# if sys.version_info >= (3, 5, 3):
#     from typing import Awaitable
# else:
#     from typing_extensions import Awaitable
#
# if sys.version_info >= (3, 6):
#     # from os import PathLike
#     from pathlib import PurePath as PathLike
#     from typing import ContextManager
# else:
#     from pathlib import PurePath as PathLike
#
#     from typing_extensions import ContextManager
#
# if sys.version_info >= (3, 8):
#     from typing import Protocol, TypedDict
# else:
#     from typing_extensions import Protocol, TypedDict
#
# _T = TypeVar("_T")
# _F = TypeVar("_F", bound=Callable[..., Any])
# ExcInfo = Tuple[
#     Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]
# ]
#
#
# class _GeneratorContextManager(ContextManager[_T], Generic[_T]):
#     def __call__(self, func: _F) -> _F:
#         ...
#
#
# Catcher = _GeneratorContextManager[None]
# Contextualizer = _GeneratorContextManager[None]
# AwaitableCompleter = Awaitable
#
#
# class Level(NamedTuple):
#     name: str
#     no: int
#     color: str
#     icon: str
#
#
# class _RecordAttribute:
#     def __repr__(self) -> str:
#         ...
#
#     def __format__(self, spec: str) -> str:
#         ...
#
#
# class RecordFile(_RecordAttribute):
#     name: str
#     path: str
#
#
# class RecordLevel(_RecordAttribute):
#     name: str
#     no: int
#     icon: str
#
#
# class RecordThread(_RecordAttribute):
#     id: int
#     name: str
#
#
# class RecordProcess(_RecordAttribute):
#     id: int
#     name: str
#
#
# class RecordException(NamedTuple):
#     type: Optional[Type[BaseException]]
#     value: Optional[BaseException]
#     traceback: Optional[TracebackType]
#
#
# class Record(TypedDict):
#     elapsed: timedelta
#     exception: Optional[RecordException]
#     extra: Dict[Any, Any]
#     file: RecordFile
#     function: str
#     level: RecordLevel
#     line: int
#     message: str
#     module: str
#     name: Union[str, None]
#     process: RecordProcess
#     thread: RecordThread
#     time: datetime
#
#
# class Message(str):
#     record: Record
#
#
# class Writable(Protocol):
#     def write(self, message: Message) -> None:
#         ...
#
#
# FilterDict = Dict[Union[str, None], Union[str, int, bool]]
# FilterFunction = Callable[[Record], bool]
# FormatFunction = Callable[[Record], str]
# PatcherFunction = Callable[[Record], None]
# RotationFunction = Callable[[Message, TextIO], bool]
# RetentionFunction = Callable[[List[str]], None]
# CompressionFunction = Callable[[str], None]
#
# # Actually unusable because TypedDict can't allow extra keys: python/mypy#4617
# class _HandlerConfig(TypedDict, total=False):
#     sink: Union[
#         str, PathLike, TextIO, Writable, Callable[[Message], None], Handler
#     ]
#     level: Union[str, int]
#     format: Union[str, FormatFunction]
#     filter: Optional[Union[str, FilterFunction, FilterDict]]
#     colorize: Optional[bool]
#     serialize: bool
#     backtrace: bool
#     diagnose: bool
#     enqueue: bool
#     catch: bool
#
#
# class LevelConfig(TypedDict, total=False):
#     name: str
#     no: int
#     color: str
#     icon: str
#
#
# ActivationConfig = Tuple[Union[str, None], bool]
#
#
# class Logger:
#     @overload
#     def add(
#         self,
#         sink: Union[TextIO, Writable, Callable[[Message], None], Handler],
#         *,
#         level: Union[str, int] = ...,
#         format: Union[str, FormatFunction] = ...,
#         filter: Optional[Union[str, FilterFunction, FilterDict]] = ...,
#         colorize: Optional[bool] = ...,
#         serialize: bool = ...,
#         backtrace: bool = ...,
#         diagnose: bool = ...,
#         enqueue: bool = ...,
#         catch: bool = ...
#     ) -> int:
#         ...
#
#     @overload
#     def add(
#         self,
#         sink: Callable[[Message], Awaitable[None]],
#         *,
#         level: Union[str, int] = ...,
#         format: Union[str, FormatFunction] = ...,
#         filter: Optional[Union[str, FilterFunction, FilterDict]] = ...,
#         colorize: Optional[bool] = ...,
#         serialize: bool = ...,
#         backtrace: bool = ...,
#         diagnose: bool = ...,
#         enqueue: bool = ...,
#         catch: bool = ...,
#         loop: Optional[AbstractEventLoop] = ...
#     ) -> int:
#         ...
#
#     @overload
#     def add(
#         self,
#         sink: Union[str, PathLike[str]],
#         *,
#         level: Union[str, int] = ...,
#         format: Union[str, FormatFunction] = ...,
#         filter: Optional[Union[str, FilterFunction, FilterDict]] = ...,
#         colorize: Optional[bool] = ...,
#         serialize: bool = ...,
#         backtrace: bool = ...,
#         diagnose: bool = ...,
#         enqueue: bool = ...,
#         catch: bool = ...,
#         rotation: Optional[Union[str, int, time, timedelta, RotationFunction]] = ...,
#         retention: Optional[Union[str, int, timedelta, RetentionFunction]] = ...,
#         compression: Optional[Union[str, CompressionFunction]] = ...,
#         delay: bool = ...,
#         mode: str = ...,
#         buffering: int = ...,
#         encoding: str = ...,
#         **kwargs: Any
#     ) -> int:
#         ...
#
#     def remove(self, handler_id: Optional[int] = ...) -> None:
#         ...
#
#     def complete(self) -> AwaitableCompleter:
#         ...
#
#     @overload
#     def catch(
#         self,
#         exception: Union[Type[BaseException], Tuple[Type[BaseException], ...]] = ...,
#         *,
#         level: Union[str, int] = ...,
#         reraise: bool = ...,
#         onerror: Optional[Callable[[BaseException], None]] = ...,
#         exclude: Optional[
#             Union[Type[BaseException], Tuple[Type[BaseException], ...]]
#         ] = ...,
#         default: Any = ...,
#         message: str = ...
#     ) -> Catcher:
#         ...
#
#     @overload
#     def catch(self, exception: _F) -> _F:
#         ...
#
#     def opt(
#         self,
#         *,
#         exception: Optional[Union[bool, ExcInfo, BaseException]] = ...,
#         record: bool = ...,
#         lazy: bool = ...,
#         colors: bool = ...,
#         raw: bool = ...,
#         capture: bool = ...,
#         depth: int = ...,
#         ansi: bool = ...
#     ):
#         ...
#
#     def bind(__self, **kwargs: Any):
#         ...
#
#     def contextualize(__self, **kwargs: Any) -> Contextualizer:
#         ...
#
#     def patch(self, patcher: PatcherFunction):
#         ...
#
#     @overload
#     def level(self, name: str) -> Level:
#         ...
#
#     @overload
#     def level(
#         self,
#         name: str,
#         no: int = ...,
#         color: Optional[str] = ...,
#         icon: Optional[str] = ...,
#     ) -> Level:
#         ...
#
#     @overload
#     def level(
#         self,
#         name: str,
#         no: Optional[int] = ...,
#         color: Optional[str] = ...,
#         icon: Optional[str] = ...,
#     ) -> Level:
#         ...
#
#     def disable(self, name: Union[str, None]) -> None:
#         ...
#
#     def enable(self, name: Union[str, None]) -> None:
#         ...
#
#     def configure(
#         self,
#         *,
#         handlers: Sequence[Dict[str, Any]] = ...,
#         levels: Optional[Sequence[LevelConfig]] = ...,
#         extra: Optional[Dict[Any, Any]] = ...,
#         patcher: Optional[PatcherFunction] = ...,
#         activation: Optional[Sequence[ActivationConfig]] = ...
#     ) -> List[int]:
#         ...
#
#     # @staticmethod cannot be used with @overload in mypy (python/mypy#7781).
#     # However Logger is not exposed and logger is an instance of Logger
#     # so for type checkers it is all the same whether it is defined here
#     # as a static method or an instance method.
#     @overload
#     def parse(
#         self,
#         file: Union[str, PathLike[str], TextIO],
#         pattern: Union[str, Pattern[str]],
#         *,
#         cast: Union[
#             Dict[str, Callable[[str], Any]], Callable[[Dict[str, str]], None]
#         ] = ...,
#         chunk: int = ...
#     ) -> Generator[Dict[str, Any], None, None]:
#         ...
#
#     @overload
#     def parse(
#         self,
#         file: BinaryIO,
#         pattern: Union[bytes, Pattern[bytes]],
#         *,
#         cast: Union[
#             Dict[str, Callable[[bytes], Any]], Callable[[Dict[str, bytes]], None]
#         ] = ...,
#         chunk: int = ...
#     ) -> Generator[Dict[str, Any], None, None]:
#         ...
#
#     @overload
#     def trace(__self, __message: str, *args: Any, **kwargs: Any) -> None:
#         ...
#
#     @overload
#     def trace(__self, __message: Any) -> None:
#         ...
#
#     @overload
#     def debug(__self, __message: str, *args: Any, **kwargs: Any) -> None:
#         ...
#
#     @overload
#     def debug(__self, __message: Any) -> None:
#         ...
#
#     @overload
#     def info(__self, __message: str, *args: Any, **kwargs: Any) -> None:
#         ...
#
#     @overload
#     def info(__self, __message: Any) -> None:
#         ...
#
#     @overload
#     def success(__self, __message: str, *args: Any, **kwargs: Any) -> None:
#         ...
#
#     @overload
#     def success(__self, __message: Any) -> None:
#         ...
#
#     @overload
#     def warning(__self, __message: str, *args: Any, **kwargs: Any) -> None:
#         ...
#
#     @overload
#     def warning(__self, __message: Any) -> None:
#         ...
#
#     @overload
#     def error(__self, __message: str, *args: Any, **kwargs: Any) -> None:
#         ...
#
#     @overload
#     def error(__self, __message: Any) -> None:
#         ...
#
#     @overload
#     def critical(__self, __message: str, *args: Any, **kwargs: Any) -> None:
#         ...
#
#     @overload
#     def critical(__self, __message: Any) -> None:
#         ...
#
#     @overload
#     def exception(__self, __message: str, *args: Any, **kwargs: Any) -> None:
#         ...
#
#     @overload
#     def exception(__self, __message: Any) -> None:
#         ...
#
#     @overload
#     def log(
#         __self, __level: Union[int, str], __message: str, *args: Any, **kwargs: Any
#     ) -> None:
#         ...
#
#     @overload
#     def log(__self, __level: Union[int, str], __message: Any) -> None:
#         ...
#
#     def start(self, *args: Any, **kwargs: Any):
#         ...
#
#     def stop(self, *args: Any, **kwargs: Any):
#         ...
#
#
# logger: Logger

from loguru import logger



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 2:14 PM
# @File: __init__.py
# @Email: mlshenkai@163.com
from .ml_classifier import ClassicClassifier
from .text_cluster import TextCluster



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 2:14 PM
# @File: ml_classifier
# @Email: mlshenkai@163.com

import argparse
import os
import sys
import pickle
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import MultinomialNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import train_test_split
from loguru import logger

from EasyAI.TextClassifier.base_classifier import ClassifierABC, load_data
from EasyAI.TextClassifier.tokenizer import Tokenizer

pwd_path = os.path.abspath(os.path.dirname(__file__))
default_stopwords_path = os.path.join(pwd_path, "stopwords.txt")


class ClassicClassifier(ClassifierABC):
    def __init__(
        self,
        model_dir,
        model_name_or_model="lr",
        feature_name_or_feature="tfidf",
        stopwords_path=default_stopwords_path,
        tokenizer=None,
    ):
        """
        经典机器学习分类模型，支持lr, random_forest, decision_tree, knn, bayes, svm, xgboost
        @param model_dir: 模型保存路径
        @param model_name_or_model:
        @param feature_name_or_feature:
        @param stopwords_path:
        @param tokenizer: 切词器，默认为jieba切词
        """
        self.model_dir = model_dir
        if isinstance(model_name_or_model, str):
            model_name = model_name_or_model.lower()
            if model_name not in [
                "lr",
                "random_forest",
                "decision_tree",
                "knn",
                "bayes",
                "xgboost",
                "svm",
            ]:
                raise ValueError("model_name not found.")
            logger.debug(f"model_name: {model_name}")
            self.model = self.get_model(model_name)
        elif hasattr(model_name_or_model, "fit"):
            self.model = model_name_or_model
        else:
            raise ValueError("model_name_or_model set error.")
        if isinstance(feature_name_or_feature, str):
            feature_name = feature_name_or_feature.lower()
            if feature_name not in ["tfidf", "count"]:
                raise ValueError("feature_name not found.")
            logger.debug(f"feature_name: {feature_name}")
            if feature_name == "tfidf":
                self.feature = TfidfVectorizer(ngram_range=(1, 2))
            else:
                self.feature = CountVectorizer(ngram_range=(1, 2))
        elif hasattr(feature_name_or_feature, "fit_transform"):
            self.feature = feature_name_or_feature
        else:
            raise ValueError("feature_name_or_feature set error.")
        self.is_trained = False
        self.stopwords = (
            set(self.load_list(stopwords_path))
            if stopwords_path and os.path.exists(stopwords_path)
            else set()
        )
        self.tokenizer = tokenizer if tokenizer else Tokenizer()

    def __str__(self):
        return f"ClassicClassifier instance ({self.model}, stopwords size: {len(self.stopwords)})"

    @staticmethod
    def get_model(model_type):
        if model_type in ["lr", "logistic_regression"]:
            model = LogisticRegression(
                solver="lbfgs", fit_intercept=False
            )  # 快，准确率一般。val mean acc:0.91
        elif model_type == "random_forest":
            model = RandomForestClassifier(
                n_estimators=300
            )  # 速度还行，准确率一般。val mean acc:0.93125
        elif model_type == "decision_tree":
            model = DecisionTreeClassifier()  # 速度快，准确率低。val mean acc:0.62
        elif model_type == "knn":
            model = KNeighborsClassifier()  # 速度一般，准确率低。val mean acc:0.675
        elif model_type == "bayes":
            model = MultinomialNB(
                alpha=0.1, fit_prior=False
            )  # 速度快，准确率低。val mean acc:0.62
        elif model_type == "xgboost":
            from xgboost import XGBClassifier

            model = XGBClassifier()  # 速度慢，准确率高。val mean acc:0.95
        elif model_type == "svm":
            model = SVC(
                kernel="linear", probability=True
            )  # 速度慢，准确率高，val mean acc:0.945
        else:
            raise ValueError("model type set error.")
        return model

    @staticmethod
    def load_list(path):
        return [word for word in open(path, "r", encoding="utf-8").read().split()]

    def tokenize_sentences(self, sentences):
        """
        Tokenize input text
        :param sentences: list of text, eg: [text1, text2, ...]
        :return: X_tokens
        """
        X_tokens = [
            " ".join(
                [w for w in self.tokenizer.tokenize(line) if w not in self.stopwords]
            )
            for line in sentences
        ]
        return X_tokens

    def load_pkl(self, pkl_path):
        """
        加载词典文件
        :param pkl_path:
        :return:
        """
        with open(pkl_path, "rb") as f:
            result = pickle.load(f)
        return result

    def save_pkl(self, vocab, pkl_path, overwrite=True):
        """
        存储文件
        :param pkl_path:
        :param overwrite:
        :return:
        """
        if pkl_path and os.path.exists(pkl_path) and not overwrite:
            return
        if pkl_path:
            with open(pkl_path, "wb") as f:
                pickle.dump(vocab, f, protocol=pickle.HIGHEST_PROTOCOL)  # python3

    def train(
        self,
        data_list_or_path,
        header=None,
        names=("labels", "text"),
        delimiter="\t",
        test_size=0.1,
    ):
        """
        Train model with data_list_or_path and save model to model_dir
        @param data_list_or_path:
        @param header:
        @param names:
        @param delimiter:
        @param test_size:
        @return:
        """
        # load data
        X, y, data_df = load_data(
            data_list_or_path,
            header=header,
            names=names,
            delimiter=delimiter,
            is_train=True,
        )
        # split validation set
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, random_state=1
        )
        # train model
        logger.debug(f"X_train size: {len(X_train)}, X_test size: {len(X_test)}")
        assert len(X_train) == len(y_train)
        logger.debug(f"X_train sample:\n{X_train[:3]}\ny_train sample:\n{y_train[:3]}")
        logger.debug(f"num_classes:{len(set(y))}")
        # tokenize text
        X_train_tokens = self.tokenize_sentences(X_train)
        logger.debug(f"X_train_tokens sample:\n{X_train_tokens[:3]}")
        X_train_feat = self.feature.fit_transform(X_train_tokens)
        # fit
        self.model.fit(X_train_feat, y_train)
        self.is_trained = True
        # evaluate
        test_acc = self.evaluate(X_test, y_test)
        logger.debug(
            f"evaluate, X size: {len(X_test)}, y size: {len(y_test)}, acc: {test_acc}"
        )
        # save model
        self.save_model()
        return test_acc

    def predict(self, sentences: list):
        """
        Predict labels and label probability for sentences.
        @param sentences: list, input text list, eg: [text1, text2, ...]
        @return: predict_label, predict_prob
        """
        if not self.is_trained:
            raise ValueError("model not trained.")
        # tokenize text
        X_tokens = self.tokenize_sentences(sentences)
        # transform
        X_feat = self.feature.transform(X_tokens)
        predict_labels = self.model.predict(X_feat)
        probs = self.model.predict_proba(X_feat)
        predict_probs = [
            prob[np.where(self.model.classes_ == label)][0]
            for label, prob in zip(predict_labels, probs)
        ]
        return predict_labels, predict_probs

    def evaluate_model(
        self, data_list_or_path, header=None, names=("labels", "text"), delimiter="\t"
    ):
        X_test, y_test, df = load_data(
            data_list_or_path, header=header, names=names, delimiter=delimiter
        )
        return self.evaluate(X_test, y_test)

    def evaluate(self, X_test, y_test):
        """
        Evaluate model.
        @param X_test:
        @param y_test:
        @return: accuracy score
        """
        if not self.is_trained:
            raise ValueError("model not trained.")
        # evaluate the model
        y_pred, _ = self.predict(X_test)
        acc_score = metrics.accuracy_score(y_test, y_pred)
        return acc_score

    def load_model(self):
        """
        Load model from model_dir
        @return:
        """
        model_path = os.path.join(self.model_dir, "classifier_model.pkl")
        if os.path.exists(model_path):
            self.model = self.load_pkl(model_path)
            feature_path = os.path.join(self.model_dir, "classifier_feature.pkl")
            self.feature = self.load_pkl(feature_path)
            logger.info(f"Loaded model: {model_path}.")
            self.is_trained = True
        else:
            logger.error(f"{model_path} not exists.")
            self.is_trained = False
        return self.is_trained

    def save_model(self):
        """
        Save model to model_dir
        @return:
        """
        if self.model_dir:
            os.makedirs(self.model_dir, exist_ok=True)
        if self.is_trained:
            feature_path = os.path.join(self.model_dir, "classifier_feature.pkl")
            self.save_pkl(self.feature, feature_path)
            model_path = os.path.join(self.model_dir, "classifier_model.pkl")
            self.save_pkl(self.model, model_path)
            logger.info(f"Saved model: {model_path}, feature_path: {feature_path}")
        else:
            logger.error("model is not trained, please train model first")
        return self.model, self.feature



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/23 4:38 PM
# @File: bert_classifer
# @Email: mlshenkai@163.com

import argparse
import json
import os

import numpy as np
import torch
from loguru import logger
from sklearn.model_selection import train_test_split
from EasyAI.TextClassifier.base_classifier import ClassifierABC, load_data
from EasyAI.TextClassifier.data_helper import set_seed
from EasyAI.TextClassifier.transformer.models.bert_classification_model import (
    BertClassificationModel,
    BertClassificationArgs,
)

pwd_path = os.path.abspath(os.path.dirname(__file__))
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
use_cuda = torch.cuda.is_available()


class BertClassifier(ClassifierABC):
    def __init__(
        self,
        model_dir,
        num_classes,
        model_type="bert",
        model_name="bert-base-chinese",
        num_epochs=3,
        batch_size=64,
        max_seq_length=128,
        multi_label=False,
        labels_sep=",",
        args=None,
    ):
        """
        Init classification model
        @param model_dir: str, model dir
        @param model_type: support 'bert', 'albert', 'roberta', 'xlnet'
        @param model_name:
        @param num_classes:
        @param num_epochs:
        @param batch_size:
        @param max_seq_length:
        @param multi_label:
        @param labels_sep:
        @param args:
        """
        default_args = {
            "output_dir": model_dir,
            "max_seq_length": max_seq_length,
            "num_train_epochs": num_epochs,
            "train_batch_size": batch_size,
            "best_model_dir": os.path.join(model_dir, "best_model"),
            "labels_sep": labels_sep,
        }
        train_args = BertClassificationArgs()
        if args and isinstance(args, dict):
            train_args.update_from_dict(args)
        train_args.update_from_dict(default_args)

        self.model = BertClassificationModel(
            model_type=model_type,
            model_name=model_name,
            num_labels=num_classes,
            multi_label=multi_label,
            args=train_args,
            use_cuda=use_cuda,
        )
        self.model_dir = model_dir
        self.model_type = model_type
        self.num_classes = num_classes
        self.batch_size = batch_size
        self.max_seq_length = max_seq_length
        self.num_epochs = num_epochs
        self.train_args = train_args
        self.use_cuda = use_cuda
        self.multi_label = multi_label
        self.labels_sep = labels_sep
        self.label_vocab_path = os.path.join(self.model_dir, "label_vocab.json")
        self.is_trained = False

    def __str__(self):
        return f"BertClassifier instance ({self.model})"

    def train(
        self,
        data_list_or_path,
        dev_data_list_or_path=None,
        header=None,
        names=("labels", "text"),
        delimiter="\t",
        test_size=0.1,
    ):
        """
        Train model with data_list_or_path and save model to model_dir
        @param data_list_or_path:
        @param dev_data_list_or_path:
        @param header:
        @param names:
        @param delimiter:
        @param test_size:
        @return:
        """
        logger.debug("train model ...")
        SEED = 1
        set_seed(SEED)
        # load data
        X, y, data_df = load_data(
            data_list_or_path,
            header=header,
            names=names,
            delimiter=delimiter,
            labels_sep=self.labels_sep,
            is_train=True,
        )
        if self.model_dir:
            os.makedirs(self.model_dir, exist_ok=True)
        labels_map = self.build_labels_map(
            y, self.label_vocab_path, self.multi_label, self.labels_sep
        )
        labels_list = sorted(list(labels_map.keys()))
        if dev_data_list_or_path is not None:
            dev_X, dev_y, dev_df = load_data(
                dev_data_list_or_path,
                header=header,
                names=names,
                delimiter=delimiter,
                labels_sep=self.labels_sep,
                is_train=False,
            )
            train_data = data_df
            dev_data = dev_df
        else:
            if test_size > 0:
                train_data, dev_data = train_test_split(
                    data_df, test_size=test_size, random_state=SEED
                )
            else:
                train_data = data_df
                dev_data = None
        logger.debug(f"train_data size: {len(train_data)}")
        logger.debug(f"train_data sample:\n{train_data[:3]}")
        if dev_data is not None and dev_data.size:
            logger.debug(f"dev_data size: {len(dev_data)}")
            logger.debug(f"dev_data sample:\n{dev_data[:3]}")
        # train model
        if self.train_args.lazy_loading:
            train_data = data_list_or_path
            dev_data = dev_data_list_or_path
        if dev_data is not None:
            self.model.train_model(
                train_data,
                eval_df=dev_data,
                args={"labels_map": labels_map, "labels_list": labels_list},
            )
        else:
            self.model.train_model(
                train_data, args={"labels_map": labels_map, "labels_list": labels_list}
            )
        self.is_trained = True
        logger.debug("train model done")

    def predict(self, sentences: list, top_k: int = None):
        """
        Predict labels and label probability for sentences.
        @param sentences: list, input text list, eg: [text1, text2, ...]
        @param top_k: int, if not None return top_k
        @return: predict_labels, predict_probs
        """
        if not self.is_trained:
            raise ValueError("model not trained.")
        # predict
        predictions, raw_outputs = self.model.predict(sentences)
        if self.multi_label:
            return predictions, raw_outputs
        else:
            # predict probability
            predict_probs = [
                1 - np.exp(-np.max(raw_output))
                for raw_output, prediction in zip(raw_outputs, predictions)
            ]
            return predictions, predict_probs

    def evaluate_model(
        self,
        data_list_or_path,
        header=None,
        names=("labels", "text"),
        delimiter="\t",
        **kwargs,
    ):
        if self.train_args.lazy_loading:
            eval_df = data_list_or_path
        else:
            X_test, y_test, eval_df = load_data(
                data_list_or_path,
                header=header,
                names=names,
                delimiter=delimiter,
                labels_sep=self.labels_sep,
            )
        if not self.is_trained:
            self.load_model()
        result, model_outputs, wrong_predictions = self.model.eval_model(
            eval_df,
            output_dir=self.model_dir,
            **kwargs,
        )
        return result

    def load_model(self):
        """
        Load model from model_dir
        @return:
        """
        model_path = os.path.join(self.model_dir, "pytorch_model.bin")
        if os.path.exists(model_path):
            labels_map = json.load(open(self.label_vocab_path, "r", encoding="utf-8"))
            labels_list = sorted(list(labels_map.keys()))
            num_classes = len(labels_map)
            assert (
                num_classes == self.num_classes
            ), f"num_classes not match, {num_classes} != {self.num_classes}"
            self.train_args.update_from_dict(
                {"labels_map": labels_map, "labels_list": labels_list}
            )
            self.model = BertClassificationModel(
                model_type=self.model_type,
                model_name=self.model_dir,
                num_labels=self.num_classes,
                multi_label=self.multi_label,
                args=self.train_args,
                use_cuda=self.use_cuda,
            )
            self.is_trained = True
        else:
            logger.error(f"{model_path} not exists.")
            self.is_trained = False
        return self.is_trained

    @staticmethod
    def build_labels_map(y, label_vocab_path, multi_label=False, labels_sep=","):
        """
        Build labels map
        @param y:
        @param label_vocab_path:
        @param multi_label:
        @param labels_sep:
        @return:
        """
        if multi_label:
            labels = set()
            for label in y.tolist():
                if isinstance(label, str):
                    labels.update(label.split(labels_sep))
                elif isinstance(label, list):
                    labels.update(range(len(label)))
                else:
                    labels.add(label)
        else:
            labels = set(y.tolist())
        labels = sorted(list(labels))
        id_label_map = {id: v for id, v in enumerate(labels)}
        label_id_map = {v: k for k, v in id_label_map.items()}
        json.dump(
            label_id_map,
            open(label_vocab_path, "w", encoding="utf-8"),
            ensure_ascii=False,
            indent=4,
        )
        logger.debug(
            f"label vocab size: {len(label_id_map)}, label_vocab_path: {label_vocab_path}"
        )
        return label_id_map


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Bert Text Classification")
    parser.add_argument(
        "--pretrain_model_type",
        default="bert",
        type=str,
        help="pretrained huggingface model type",
    )
    parser.add_argument(
        "--pretrain_model_name",
        default="bert-base-chinese",
        type=str,
        help="pretrained huggingface model name",
    )
    parser.add_argument(
        "--model_dir", default="models/bert", type=str, help="save model dir"
    )
    parser.add_argument(
        "--data_path",
        default=os.path.join(pwd_path, "../examples/thucnews_train_1w.txt"),
        type=str,
        help="sample data file path",
    )
    parser.add_argument("--num_classes", default=10, type=int, help="number of classes")
    parser.add_argument("--num_epochs", default=3, type=int, help="train epochs")
    parser.add_argument("--batch_size", default=64, type=int, help="train batch size")
    parser.add_argument(
        "--max_seq_length",
        default=128,
        type=int,
        help="max seq length, trim longer sentence.",
    )
    args = parser.parse_args()
    print(args)
    # create model
    m = BertClassifier(
        model_dir=args.model_dir,
        num_classes=args.num_classes,
        model_type=args.pretrain_model_type,
        model_name=args.pretrain_model_name,
        num_epochs=args.num_epochs,
        batch_size=args.batch_size,
        max_seq_length=args.max_seq_length,
        multi_label=False,
    )
    # train model
    m.train(data_list_or_path=args.data_path)
    # load trained best model and predict
    m.load_model()
    print("best model loaded from file, and predict")
    X, y, _ = load_data(args.data_path)
    X = X[:5]
    y = y[:5]
    predict_labels, predict_probs = m.predict(X)
    for text, pred_label, pred_prob, y_truth in zip(
        X, predict_labels, predict_probs, y
    ):
        print(text, "pred:", pred_label, pred_prob, " truth:", y_truth)



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/23 4:22 PM
# @File: bert_classification_utils
# @Email: mlshenkai@163.com
import json
import linecache
import os
import sys
import warnings

try:
    from collections import Iterable, Mapping
except ImportError:
    from collections.abc import Iterable, Mapping
from dataclasses import asdict, dataclass, field
from io import open
from multiprocessing import Pool
from multiprocessing import cpu_count
from numbers import Real
from typing import Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from datasets import Dataset as HFDataset
from datasets import load_dataset
from torch.utils.data import Dataset
from tqdm.auto import tqdm
from loguru import logger


def get_default_process_count():
    process_count = cpu_count() - 2 if cpu_count() > 2 else 1
    if sys.platform == "win32":
        process_count = min(process_count, 61)

    return process_count


def get_special_tokens():
    return ["<s>", "<pad>", "</s>", "<unk>", "<mask>"]


@dataclass
class ModelArgs:
    adafactor_beta1: float = None
    adafactor_clip_threshold: float = 1.0
    adafactor_decay_rate: float = -0.8
    adafactor_eps: tuple = field(default_factory=lambda: (1e-30, 1e-3))
    adafactor_relative_step: bool = True
    adafactor_scale_parameter: bool = True
    adafactor_warmup_init: bool = True
    adam_epsilon: float = 1e-8
    best_model_dir: str = "outputs/best_model"
    cache_dir: str = "cache_dir/"
    config: dict = field(default_factory=dict)
    cosine_schedule_num_cycles: float = 0.5
    custom_layer_parameters: list = field(default_factory=list)
    custom_parameter_groups: list = field(default_factory=list)
    dataloader_num_workers: int = 0
    do_lower_case: bool = False
    dynamic_quantize: bool = False
    early_stopping_consider_epochs: bool = False
    early_stopping_delta: float = 0
    early_stopping_metric: str = "eval_loss"
    early_stopping_metric_minimize: bool = True
    early_stopping_patience: int = 3
    encoding: str = None
    eval_batch_size: int = 8
    evaluate_during_training: bool = True
    evaluate_during_training_silent: bool = True
    evaluate_during_training_steps: int = 2000
    evaluate_during_training_verbose: bool = False
    evaluate_each_epoch: bool = True
    fp16: bool = False
    gradient_accumulation_steps: int = 1
    learning_rate: float = 4e-5
    local_rank: int = -1
    logging_steps: int = 50
    loss_type: str = None
    loss_args: dict = field(default_factory=dict)
    manual_seed: int = None
    max_grad_norm: float = 1.0
    max_seq_length: int = 128
    model_name: str = None
    model_type: str = None
    multiprocessing_chunksize: int = -1
    n_gpu: int = 1
    no_cache: bool = True
    no_save: bool = False
    not_saved_args: list = field(default_factory=list)
    num_train_epochs: int = 1
    optimizer: str = "AdamW"
    output_dir: str = "outputs/"
    overwrite_output_dir: bool = True
    polynomial_decay_schedule_lr_end: float = 1e-7
    polynomial_decay_schedule_power: float = 1.0
    process_count: int = field(default_factory=get_default_process_count)
    quantized_model: bool = False
    reprocess_input_data: bool = True
    save_best_model: bool = True
    save_eval_checkpoints: bool = True
    save_model_every_epoch: bool = True
    save_optimizer_and_scheduler: bool = True
    save_steps: int = 2000
    scheduler: str = "linear_schedule_with_warmup"
    silent: bool = False
    skip_special_tokens: bool = True
    tensorboard_dir: str = None
    thread_count: int = None
    tokenizer_name: str = None
    tokenizer_type: str = None
    train_batch_size: int = 8
    train_custom_parameters_only: bool = False
    use_cached_eval_features: bool = False
    use_early_stopping: bool = True
    use_hf_datasets: bool = False
    use_multiprocessing: bool = False
    use_multiprocessing_for_evaluation: bool = False
    wandb_kwargs: dict = field(default_factory=dict)
    wandb_project: str = None
    warmup_ratio: float = 0.06
    warmup_steps: int = 0
    weight_decay: float = 0.0

    def update_from_dict(self, new_values):
        if isinstance(new_values, dict):
            for key, value in new_values.items():
                setattr(self, key, value)
        else:
            raise (TypeError(f"{new_values} is not a Python dict."))

    def get_args_for_saving(self):
        args_for_saving = {
            key: value
            for key, value in asdict(self).items()
            if key not in self.not_saved_args
        }
        if "settings" in args_for_saving["wandb_kwargs"]:
            del args_for_saving["wandb_kwargs"]["settings"]
        return args_for_saving

    def save(self, output_dir):
        os.makedirs(output_dir, exist_ok=True)
        with open(os.path.join(output_dir, "model_args.json"), "w") as f:
            args_dict = self.get_args_for_saving()
            if args_dict["tokenizer_type"] is not None and not isinstance(
                args_dict["tokenizer_type"], str
            ):
                args_dict["tokenizer_type"] = type(args_dict["tokenizer_type"]).__name__
            json.dump(args_dict, f)

    def load(self, input_dir):
        if input_dir:
            model_args_file = os.path.join(input_dir, "model_args.json")
            if os.path.isfile(model_args_file):
                with open(model_args_file, "r") as f:
                    model_args = json.load(f)

                self.update_from_dict(model_args)


@dataclass
class BertClassificationArgs(ModelArgs):
    """
    Model args for a ClassificationModel
    """

    model_class: str = "ClassificationModel"
    labels_list: list = field(default_factory=list)
    labels_map: dict = field(default_factory=dict)
    lazy_delimiter: str = "\t"
    lazy_labels_column: int = 1
    lazy_loading: bool = False
    lazy_loading_start_line: int = 1
    lazy_text_a_column: bool = None
    lazy_text_b_column: bool = None
    lazy_text_column: int = 0
    onnx: bool = False
    regression: bool = False
    sliding_window: bool = False
    special_tokens_list: list = field(default_factory=list)
    stride: float = 0.8
    tie_value: int = 1
    threshold: float = 0.5


class InputExample(object):
    """A single training/test example for simple sequence classification."""

    def __init__(self, guid, text_a, text_b=None, label=None):
        """
        Constructs a InputExample.

        Args:
            guid: Unique id for the example.
            text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
            text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
            label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
        """

        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.label = label

    def __repr__(self):
        return str(
            {
                "guid": self.guid,
                "text_a": self.text_a,
                "text_b": self.text_b,
                "label": self.label,
            }
        )


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, input_ids, input_mask, segment_ids, label_id):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.label_id = label_id


def preprocess_data_multiprocessing(data):
    text_a, text_b, tokenizer, max_seq_length = data

    examples = tokenizer(
        text=text_a,
        text_pair=text_b,
        max_length=max_seq_length,
        truncation=True,
        padding="max_length",
        return_tensors="pt",
    )

    return examples


def preprocess_batch_for_hf_dataset(dataset, tokenizer, max_seq_length):
    if "text_b" in dataset:
        return tokenizer(
            text=dataset["text_a"],
            text_pair=dataset["text_b"],
            truncation=True,
            padding="max_length",
            max_length=max_seq_length,
        )
    else:
        return tokenizer(
            text=dataset["text"],
            truncation=True,
            padding="max_length",
            max_length=max_seq_length,
        )


def preprocess_data(text_a, text_b, labels, tokenizer, max_seq_length):
    return tokenizer(
        text=text_a,
        text_pair=text_b,
        truncation=True,
        padding="max_length",
        max_length=max_seq_length,
        return_tensors="pt",
    )


def build_classification_dataset(
    data, tokenizer, args, mode, multi_label, output_mode, no_cache
):
    cached_features_file = os.path.join(
        args.cache_dir,
        "cached_{}_{}_{}_{}_{}".format(
            mode,
            args.model_type,
            args.max_seq_length,
            len(args.labels_list),
            len(data),
        ),
    )

    if os.path.exists(cached_features_file) and (
        (not args.reprocess_input_data and not args.no_cache)
        or (mode == "dev" and args.use_cached_eval_features and not args.no_cache)
    ):
        data = torch.load(cached_features_file)
        logger.info(f" Features loaded from cache at {cached_features_file}")
        examples, labels = data
    else:
        logger.debug(" Converting to features started. Cache is not used.")

        if len(data) == 3:
            # Sentence pair task
            text_a, text_b, labels = data
        else:
            text_a, labels = data
            text_b = None

        # If labels_map is defined, then labels need to be replaced with ints
        if args.labels_map and not args.regression:
            if multi_label:
                if isinstance(labels[0], str):
                    labels = [
                        [
                            int(1) if i in label.split(args.labels_sep) else int(0)
                            for i in args.labels_list
                        ]
                        for label in labels
                    ]
                elif isinstance(labels[0], list) and isinstance(labels[0][0], int):
                    labels = labels
                else:
                    labels = [[args.labels_map[l] for l in label] for label in labels]
            else:
                labels = [args.labels_map[label] for label in labels]

        if (mode == "train" and args.use_multiprocessing) or (
            mode == "dev" and args.use_multiprocessing_for_evaluation
        ):
            if args.multiprocessing_chunksize == -1:
                chunksize = max(len(data) // (args.process_count * 2), 500)
            else:
                chunksize = args.multiprocessing_chunksize

            if text_b is not None:
                data = [
                    (
                        text_a[i : i + chunksize],
                        text_b[i : i + chunksize],
                        tokenizer,
                        args.max_seq_length,
                    )
                    for i in range(0, len(text_a), chunksize)
                ]
            else:
                data = [
                    (text_a[i : i + chunksize], None, tokenizer, args.max_seq_length)
                    for i in range(0, len(text_a), chunksize)
                ]

            with Pool(args.process_count) as p:
                examples = list(
                    tqdm(
                        p.imap(preprocess_data_multiprocessing, data),
                        total=len(text_a),
                        disable=args.silent,
                    )
                )

            examples = {
                key: torch.cat([example[key] for example in examples])
                for key in examples[0]
            }
        else:
            examples = preprocess_data(
                text_a, text_b, labels, tokenizer, args.max_seq_length
            )

        if output_mode == "classification":
            labels = torch.tensor(labels, dtype=torch.long)
        elif output_mode == "regression":
            labels = torch.tensor(labels, dtype=torch.float)

        data = (examples, labels)

        if not args.no_cache and not no_cache:
            logger.info(" Saving features into cached file %s" % cached_features_file)
            torch.save(data, cached_features_file)

    return (examples, labels)


class ClassificationDataset(Dataset):
    def __init__(self, data, tokenizer, args, mode, multi_label, output_mode, no_cache):
        self.examples, self.labels = build_classification_dataset(
            data, tokenizer, args, mode, multi_label, output_mode, no_cache
        )

    def __len__(self):
        return len(self.examples["input_ids"])

    def __getitem__(self, index):
        return (
            {key: self.examples[key][index] for key in self.examples},
            self.labels[index],
        )


def map_labels_to_numeric(example, multi_label, args):
    if multi_label:
        if isinstance(example["labels"][0], str):
            example["labels"] = [
                int(1) if i in example["labels"].split(args.labels_sep) else int(0)
                for i in args.labels_list
            ]
        else:
            example["labels"] = [args.labels_map[label] for label in example["labels"]]
    else:
        example["labels"] = args.labels_map[example["labels"]]

    return example


def load_hf_dataset(data, tokenizer, args, multi_label):
    if isinstance(data, str):
        dataset = load_dataset(
            "csv",
            data_files=data,
            delimiter="\t",
            download_mode="force_redownload"
            if args.reprocess_input_data
            else "reuse_dataset_if_exists",
        )
    else:
        dataset = HFDataset.from_pandas(data)

    if args.labels_map and not args.regression:
        dataset = dataset.map(lambda x: map_labels_to_numeric(x, multi_label, args))

    dataset = dataset.map(
        lambda x: preprocess_batch_for_hf_dataset(
            x, tokenizer=tokenizer, max_seq_length=args.max_seq_length
        ),
        batched=True,
    )

    if args.model_type in ["bert", "xlnet", "albert", "layoutlm"]:
        dataset.set_format(
            type="pt",
            columns=["input_ids", "token_type_ids", "attention_mask", "labels"],
        )
    else:
        dataset.set_format(type="pt", columns=["input_ids", "attention_mask", "labels"])

    if isinstance(data, str):
        # This is not necessarily a train dataset. The datasets library insists on calling it train.
        return dataset["train"]
    else:
        return dataset


def convert_example_to_feature(
    example_row,
    pad_token=0,
    sequence_a_segment_id=0,
    sequence_b_segment_id=1,
    cls_token_segment_id=1,
    pad_token_segment_id=0,
    mask_padding_with_zero=True,
    sep_token_extra=False,
):
    (
        example,
        max_seq_length,
        tokenizer,
        output_mode,
        cls_token_at_end,
        cls_token,
        sep_token,
        cls_token_segment_id,
        pad_on_left,
        pad_token_segment_id,
        sep_token_extra,
        multi_label,
        stride,
        pad_token,
        add_prefix_space,
        pad_to_max_length,
    ) = example_row

    if add_prefix_space and not example.text_a.startswith(" "):
        tokens_a = tokenizer.tokenize(" " + example.text_a)
    else:
        tokens_a = tokenizer.tokenize(example.text_a)

    tokens_b = None
    if example.text_b:
        if add_prefix_space and not example.text_b.startswith(" "):
            tokens_b = tokenizer.tokenize(" " + example.text_b)
        else:
            tokens_b = tokenizer.tokenize(example.text_b)
        # Modifies `tokens_a` and `tokens_b` in place so that the total
        # length is less than the specified length.
        # Account for [CLS], [SEP], [SEP] with "- 3". " -4" for RoBERTa.
        special_tokens_count = 4 if sep_token_extra else 3
        _truncate_seq_pair(tokens_a, tokens_b, max_seq_length - special_tokens_count)
    else:
        # Account for [CLS] and [SEP] with "- 2" and with "- 3" for RoBERTa.
        special_tokens_count = 3 if sep_token_extra else 2
        if len(tokens_a) > max_seq_length - special_tokens_count:
            tokens_a = tokens_a[: (max_seq_length - special_tokens_count)]

    # The convention in BERT is:
    # (a) For sequence pairs:
    #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
    #  type_ids:   0   0  0    0    0     0       0   0   1  1  1  1   1   1
    # (b) For single sequences:
    #  tokens:   [CLS] the dog is hairy . [SEP]
    #  type_ids:   0   0   0   0  0     0   0
    #
    # Where "type_ids" are used to indicate whether this is the first
    # sequence or the second sequence. The embedding vectors for `type=0` and
    # `type=1` were learned during pre-training and are added to the wordpiece
    # embedding vector (and position vector). This is not *strictly* necessary
    # since the [SEP] token unambiguously separates the sequences, but it makes
    # it easier for the model to learn the concept of sequences.
    #
    # For classification tasks, the first vector (corresponding to [CLS]) is
    # used as as the "sentence vector". Note that this only makes sense because
    # the entire model is fine-tuned.
    tokens = tokens_a + [sep_token]
    segment_ids = [sequence_a_segment_id] * len(tokens)

    if tokens_b:
        if sep_token_extra:
            tokens += [sep_token]
            segment_ids += [sequence_b_segment_id]

        tokens += tokens_b + [sep_token]

        segment_ids += [sequence_b_segment_id] * (len(tokens_b) + 1)

    if cls_token_at_end:
        tokens = tokens + [cls_token]
        segment_ids = segment_ids + [cls_token_segment_id]
    else:
        tokens = [cls_token] + tokens
        segment_ids = [cls_token_segment_id] + segment_ids

    input_ids = tokenizer.convert_tokens_to_ids(tokens)

    # The mask has 1 for real tokens and 0 for padding tokens. Only real
    # tokens are attended to.
    input_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)

    # Zero-pad up to the sequence length.
    if pad_to_max_length:
        padding_length = max_seq_length - len(input_ids)
        if pad_on_left:
            input_ids = ([pad_token] * padding_length) + input_ids
            input_mask = (
                [0 if mask_padding_with_zero else 1] * padding_length
            ) + input_mask
            segment_ids = ([pad_token_segment_id] * padding_length) + segment_ids
        else:
            input_ids = input_ids + ([pad_token] * padding_length)
            input_mask = input_mask + (
                [0 if mask_padding_with_zero else 1] * padding_length
            )
            segment_ids = segment_ids + ([pad_token_segment_id] * padding_length)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length

    return InputFeatures(
        input_ids=input_ids,
        input_mask=input_mask,
        segment_ids=segment_ids,
        label_id=example.label,
    )


def convert_example_to_feature_sliding_window(
    example_row,
    pad_token=0,
    sequence_a_segment_id=0,
    sequence_b_segment_id=1,
    cls_token_segment_id=1,
    pad_token_segment_id=0,
    mask_padding_with_zero=True,
    sep_token_extra=False,
):
    (
        example,
        max_seq_length,
        tokenizer,
        output_mode,
        cls_token_at_end,
        cls_token,
        sep_token,
        cls_token_segment_id,
        pad_on_left,
        pad_token_segment_id,
        sep_token_extra,
        multi_label,
        stride,
        pad_token,
        add_prefix_space,
        pad_to_max_length,
    ) = example_row

    if stride < 1:
        stride = int(max_seq_length * stride)

    bucket_size = max_seq_length - (3 if sep_token_extra else 2)
    token_sets = []

    if add_prefix_space and not example.text_a.startswith(" "):
        tokens_a = tokenizer.tokenize(" " + example.text_a)
    else:
        tokens_a = tokenizer.tokenize(example.text_a)

    if len(tokens_a) > bucket_size:
        token_sets = [
            tokens_a[i : i + bucket_size] for i in range(0, len(tokens_a), stride)
        ]
    else:
        token_sets.append(tokens_a)

    if example.text_b:
        raise ValueError(
            "Sequence pair tasks not implemented for sliding window tokenization."
        )

    # The convention in BERT is:
    # (a) For sequence pairs:
    #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
    #  type_ids:   0   0  0    0    0     0       0   0   1  1  1  1   1   1
    # (b) For single sequences:
    #  tokens:   [CLS] the dog is hairy . [SEP]
    #  type_ids:   0   0   0   0  0     0   0
    #
    # Where "type_ids" are used to indicate whether this is the first
    # sequence or the second sequence. The embedding vectors for `type=0` and
    # `type=1` were learned during pre-training and are added to the wordpiece
    # embedding vector (and position vector). This is not *strictly* necessary
    # since the [SEP] token unambiguously separates the sequences, but it makes
    # it easier for the model to learn the concept of sequences.
    #
    # For classification tasks, the first vector (corresponding to [CLS]) is
    # used as as the "sentence vector". Note that this only makes sense because
    # the entire model is fine-tuned.

    input_features = []
    for tokens_a in token_sets:
        tokens = tokens_a + [sep_token]
        segment_ids = [sequence_a_segment_id] * len(tokens)

        if cls_token_at_end:
            tokens = tokens + [cls_token]
            segment_ids = segment_ids + [cls_token_segment_id]
        else:
            tokens = [cls_token] + tokens
            segment_ids = [cls_token_segment_id] + segment_ids

        input_ids = tokenizer.convert_tokens_to_ids(tokens)

        # The mask has 1 for real tokens and 0 for padding tokens. Only real
        # tokens are attended to.
        input_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)

        # Zero-pad up to the sequence length.
        padding_length = max_seq_length - len(input_ids)
        if pad_on_left:
            input_ids = ([pad_token] * padding_length) + input_ids
            input_mask = (
                [0 if mask_padding_with_zero else 1] * padding_length
            ) + input_mask
            segment_ids = ([pad_token_segment_id] * padding_length) + segment_ids
        else:
            input_ids = input_ids + ([pad_token] * padding_length)
            input_mask = input_mask + (
                [0 if mask_padding_with_zero else 1] * padding_length
            )
            segment_ids = segment_ids + ([pad_token_segment_id] * padding_length)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length

        input_features.append(
            InputFeatures(
                input_ids=input_ids,
                input_mask=input_mask,
                segment_ids=segment_ids,
                label_id=example.label,
            )
        )

    return input_features


def convert_examples_to_features(
    examples,
    max_seq_length,
    tokenizer,
    output_mode,
    cls_token_at_end=False,
    sep_token_extra=False,
    pad_on_left=False,
    cls_token="[CLS]",
    sep_token="[SEP]",
    pad_token=0,
    sequence_a_segment_id=0,
    sequence_b_segment_id=1,
    cls_token_segment_id=1,
    pad_token_segment_id=0,
    mask_padding_with_zero=True,
    process_count=cpu_count() - 2,
    multi_label=False,
    silent=False,
    use_multiprocessing=True,
    sliding_window=False,
    flatten=False,
    stride=None,
    add_prefix_space=False,
    pad_to_max_length=True,
    args=None,
):
    """Loads a data file into a list of `InputBatch`s
    `cls_token_at_end` define the location of the CLS token:
        - False (Default, BERT/XLM pattern): [CLS] + A + [SEP] + B + [SEP]
        - True (XLNet/GPT pattern): A + [SEP] + B + [SEP] + [CLS]
    `cls_token_segment_id` define the segment id associated to the CLS token (0 for BERT, 2 for XLNet)
    """

    examples = [
        (
            example,
            max_seq_length,
            tokenizer,
            output_mode,
            cls_token_at_end,
            cls_token,
            sep_token,
            cls_token_segment_id,
            pad_on_left,
            pad_token_segment_id,
            sep_token_extra,
            multi_label,
            stride,
            pad_token,
            add_prefix_space,
            pad_to_max_length,
        )
        for example in examples
    ]

    if use_multiprocessing:
        if args.multiprocessing_chunksize == -1:
            chunksize = max(len(examples) // (args.process_count * 2), 500)
        else:
            chunksize = args.multiprocessing_chunksize
        if sliding_window:
            with Pool(process_count) as p:
                features = list(
                    tqdm(
                        p.imap(
                            convert_example_to_feature_sliding_window,
                            examples,
                            chunksize=chunksize,
                        ),
                        total=len(examples),
                        disable=silent,
                    )
                )
            if flatten:
                features = [
                    feature for feature_set in features for feature in feature_set
                ]
        else:
            with Pool(process_count) as p:
                features = list(
                    tqdm(
                        p.imap(
                            convert_example_to_feature, examples, chunksize=chunksize
                        ),
                        total=len(examples),
                        disable=silent,
                    )
                )
    else:
        if sliding_window:
            features = [
                convert_example_to_feature_sliding_window(example)
                for example in tqdm(examples, disable=silent)
            ]
            if flatten:
                features = [
                    feature for feature_set in features for feature in feature_set
                ]
        else:
            features = [
                convert_example_to_feature(example)
                for example in tqdm(examples, disable=silent)
            ]

    return features


def _truncate_seq_pair(tokens_a, tokens_b, max_length):
    """Truncates a sequence pair in place to the maximum length."""

    # This is a simple heuristic which will always truncate the longer sequence
    # one token at a time. This makes more sense than truncating an equal percent
    # of tokens from each, since if one sequence is very short then each token
    # that's truncated likely contains more information than a longer sequence.

    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_length:
            break
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()


class LazyClassificationDataset(Dataset):
    def __init__(self, data_file, tokenizer, args):
        self.data_file = data_file
        self.start_row = args.lazy_loading_start_line
        self.num_entries = self._get_n_lines(self.data_file, self.start_row)
        self.tokenizer = tokenizer
        self.args = args
        self.delimiter = args.lazy_delimiter
        if args.lazy_text_a_column is not None and args.lazy_text_b_column is not None:
            self.text_a_column = args.lazy_text_a_column
            self.text_b_column = args.lazy_text_b_column
            self.text_column = None
        else:
            self.text_column = args.lazy_text_column
            self.text_a_column = None
            self.text_b_column = None
        self.labels_column = args.lazy_labels_column

    @staticmethod
    def _get_n_lines(data_file, start_row):
        with open(data_file, encoding="utf-8") as f:
            for line_idx, _ in enumerate(f, 1):
                pass

        return line_idx - start_row

    def __getitem__(self, idx):
        line = (
            linecache.getline(self.data_file, idx + 1 + self.start_row)
            .rstrip("\n")
            .split(self.delimiter)
        )

        if not self.text_a_column and not self.text_b_column:
            text = line[self.text_column]
            label = line[self.labels_column]

            # If labels_map is defined, then labels need to be replaced with ints
            if self.args.labels_map:
                if self.args.multi_label:
                    if isinstance(label[0], str):
                        label = [
                            int(1) if i in label.split(self.args.labels_sep) else int(0)
                            for i in self.args.labels_map.keys()
                        ]
                    else:
                        label = [self.args.labels_map[l] for l in label]
                else:
                    label = self.args.labels_map[label]
            if self.args.regression:
                label = torch.tensor(float(label), dtype=torch.float)
            elif self.args.multi_label:
                label = torch.tensor([int(l) for l in label], dtype=torch.long)
            else:
                label = torch.tensor(int(label), dtype=torch.long)

            return (
                self.tokenizer(
                    text=text,
                    truncation=True,
                    padding="max_length",
                    max_length=self.args.max_seq_length,
                    return_tensors="pt",
                ),
                label,
            )
        else:
            text_a = line[self.text_a_column]
            text_b = line[self.text_b_column]
            label = line[self.labels_column]
            if self.args.regression:
                label = torch.tensor(float(label), dtype=torch.float)
            else:
                label = torch.tensor(int(label), dtype=torch.long)

            return (
                self.tokenizer(
                    text_a,
                    text_pair=text_b,
                    padding="max_length",
                    max_length=self.args.max_seq_length,
                    return_tensors="pt",
                ),
                label,
            )

    def __len__(self):
        return self.num_entries


def flatten_results(results, parent_key="", sep="/"):
    out = []
    if isinstance(results, Mapping):
        for key, value in results.items():
            pkey = parent_key + sep + str(key) if parent_key else str(key)
            out.extend(flatten_results(value, parent_key=pkey).items())
    elif isinstance(results, Iterable):
        for key, value in enumerate(results):
            pkey = parent_key + sep + str(key) if parent_key else str(key)
            out.extend(flatten_results(value, parent_key=pkey).items())
    else:
        out.append((parent_key, results))
    return dict(out)


# based on:
# https://github.com/zhezh/focalloss/blob/master/focalloss.py
# adapted from:
# https://kornia.readthedocs.io/en/v0.1.2/_modules/torchgeometry/losses/focal.html


class FocalLoss(nn.Module):
    r"""Criterion that computes Focal loss.
    According to [1], the Focal loss is computed as follows:
    .. math::
        \text{FL}(p_t) = -\alpha_t (1 - p_t)^{\gamma} \, \text{log}(p_t)
    where:
       - :math:`p_t` is the model's estimated probability for each class.
    Arguments:
        alpha (float): Weighting factor :math:`\alpha \in [0, 1]` for one-vs-others mode (weight of negative class)
                        or :math:`\alpha_i \in \R`
                        vector of weights for each class (analogous to weight argument for CrossEntropyLoss)
        gamma (float): Focusing parameter :math:`\gamma >= 0`. When 0 is equal to CrossEntropyLoss
        reduction (Optional[str]): Specifies the reduction to apply to the
         output: ‘none’ | ‘mean’ | ‘sum’.
         ‘none’: no reduction will be applied,
         ‘mean’: the sum of the output will be divided by the number of elements
                in the output, uses geometric mean if alpha set to list of weights
         ‘sum’: the output will be summed. Default: ‘none’.
        ignore_index (Optional[int]): specifies indexes that are ignored during loss calculation
         (identical to PyTorch's CrossEntropyLoss 'ignore_index' parameter). Default: -100

    Shape:
        - Input: :math:`(N, C)` where C = number of classes.
        - Target: :math:`(N)` where each value is
          :math:`0 ≤ targets[i] ≤ C−1`.
    Examples:
        >>> C = 5  # num_classes
        >>> N = 1 # num_examples
        >>> loss = FocalLoss(alpha=0.5, gamma=2.0, reduction='mean')
        >>> input = torch.randn(N, C, requires_grad=True)
        >>> target = torch.empty(N, dtype=torch.long).random_(N)
        >>> output = loss(input, target)
        >>> output.backward()
    References:
        [1] https://arxiv.org/abs/1708.02002
    """

    def __init__(
        self,
        alpha: Optional[Union[float, Iterable]] = 0.5,
        gamma: Real = 2.0,
        reduction: str = "mean",
        ignore_index: int = -100,
        epsilon=1.0e-9,
        activation_type="softmax",
    ) -> None:
        super(FocalLoss, self).__init__()
        if (
            alpha is not None
            and not isinstance(alpha, float)
            and not isinstance(alpha, Iterable)
        ):
            raise ValueError(
                f"alpha value should be None, float value or list of real values. Got: {type(alpha)}"
            )
        self.alpha: Optional[Union[float, torch.Tensor]] = (
            alpha
            if alpha is None or isinstance(alpha, float)
            else torch.FloatTensor(alpha)
        )
        if isinstance(alpha, float) and not 0.0 <= alpha <= 1.0:
            warnings.warn("[Focal Loss] alpha value is to high must be between [0, 1]")

        self.gamma: Real = gamma
        self.reduction: str = reduction
        self.ignore_index: int = ignore_index
        self.epsilon = epsilon
        self.activation_type: str = activation_type

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the criterion.
        @param input: model's output, shape [N, C] meaning N: batch_size and C: num_classes
        @param target: ground truth, shape [N], N: batch_size
        @return: shape of [N]
        """
        if not torch.is_tensor(input):
            raise TypeError(
                "Input type is not a torch.Tensor. Got {}".format(type(input))
            )
        if input.device != target.device:
            raise ValueError(
                "input and target must be in the same device. Got: {}".format(
                    input.device, target.device
                )
            )

        # filter labels
        target = target.type(torch.long)

        if self.activation_type == "sigmoid":
            multi_hot_key = target
            logits = torch.sigmoid(input)
            zero_hot_key = 1 - multi_hot_key
            focal_loss = (
                -self.alpha
                * multi_hot_key
                * torch.pow((1 - logits), self.gamma)
                * (logits + self.epsilon).log()
            )
            focal_loss += (
                -(1 - self.alpha)
                * zero_hot_key
                * torch.pow(logits, self.gamma)
                * (1 - logits + self.epsilon).log()
            )
            weights = torch.ones_like(
                focal_loss, dtype=focal_loss.dtype, device=focal_loss.device
            )
        else:
            input_mask = target != self.ignore_index
            target = target[input_mask]
            input = input[input_mask]
            # compute softmax over the classes axis
            pt = F.softmax(input, dim=1)
            logpt = F.log_softmax(input, dim=1)

            # compute focal loss
            pt = pt.gather(1, target.unsqueeze(-1)).squeeze()
            logpt = logpt.gather(1, target.unsqueeze(-1)).squeeze()
            focal_loss = -1 * (1 - pt) ** self.gamma * logpt

            weights = torch.ones_like(
                focal_loss, dtype=focal_loss.dtype, device=focal_loss.device
            )
            if self.alpha is not None:
                if isinstance(self.alpha, float):
                    alpha = torch.tensor(self.alpha, device=input.device)
                    weights = torch.where(target > 0, 1 - alpha, alpha)
                elif torch.is_tensor(self.alpha):
                    alpha = self.alpha.to(input.device)
                    weights = alpha.gather(0, target)

        tmp_loss = focal_loss * weights
        if self.reduction == "none":
            loss = tmp_loss
        elif self.reduction == "mean":
            loss = (
                tmp_loss.sum() / weights.sum()
                if torch.is_tensor(self.alpha)
                else torch.mean(tmp_loss)
            )
        elif self.reduction == "sum":
            loss = tmp_loss.sum()
        else:
            raise NotImplementedError(
                "Invalid reduction mode: {}".format(self.reduction)
            )
        return loss


def init_loss(weight, device, args, multi_label=False):
    if weight and args.loss_type:
        warnings.warn(
            f"weight and args.loss_type parameters are set at the same time"
            f"will use weighted cross entropy loss. To use {args.loss_type} set weight to None"
        )
    if weight:
        if multi_label:
            loss_fct = nn.BCEWithLogitsLoss(weight=torch.Tensor(weight).to(device))
        else:
            loss_fct = nn.CrossEntropyLoss(weight=torch.Tensor(weight).to(device))
    elif args.loss_type:
        if args.loss_type == "focal":
            if multi_label:
                loss_fct = FocalLoss(activation_type="sigmoid", **args.loss_args)
            else:
                loss_fct = FocalLoss(activation_type="softmax", **args.loss_args)
        else:
            raise NotImplementedError(f"unknown {args.loss_type} loss function")
    else:
        loss_fct = None

    return loss_fct



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 2:15 PM
# @File: tokenizer
# @Email: mlshenkai@163.com
import re
import jieba


def tokenize_words(text):
    """Word segmentation"""
    output = []
    sentences = split_2_short_text(text, include_symbol=True)
    for sentence, idx in sentences:
        if is_any_chinese_string(sentence):
            output.extend(jieba.lcut(sentence))
        else:
            output.extend(whitespace_tokenize(sentence))
    return output


class Tokenizer(object):
    """Given Full tokenization."""

    def __init__(self, lower=True):
        self.lower = lower

    def tokenize(self, text):
        """Tokenizes a piece of text."""
        res = []
        if len(text) == 0:
            return res

        if self.lower:
            text = text.lower()
        # for the multilingual and Chinese
        res = tokenize_words(text)
        return res


def split_2_short_text(text, include_symbol=True):
    """
    长句切分为短句
    :param text: str
    :param include_symbol: bool
    :return: (sentence, idx)
    """
    re_han = re.compile("([\u4E00-\u9Fa5a-zA-Z0-9+#&]+)", re.U)
    result = []
    blocks = re_han.split(text)
    start_idx = 0
    for blk in blocks:
        if not blk:
            continue
        if include_symbol:
            result.append((blk, start_idx))
        else:
            if re_han.match(blk):
                result.append((blk, start_idx))
        start_idx += len(blk)
    return result


def is_chinese(uchar):
    """判断一个unicode是否是汉字"""
    return "\u4e00" <= uchar <= "\u9fa5"


def is_all_chinese_string(string):
    """判断是否全为汉字"""
    return all(is_chinese(c) for c in string)


def is_any_chinese_string(string):
    """判断是否有中文汉字"""
    return any(is_chinese(c) for c in string)


def whitespace_tokenize(text):
    """Runs basic whitespace cleaning and splitting on a piece of text."""
    text = text.strip()
    if not text:
        return []
    tokens = text.split()
    return tokens



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/23 4:15 PM
# @File: base_classifier
# @Email: mlshenkai@163.com

import os
import pandas as pd
from loguru import logger


def load_data(
    data_list_or_path,
    header=None,
    names=("labels", "text"),
    delimiter="\t",
    labels_sep=",",
    is_train=False,
):
    """
    Encoding data_list text
    @param data_list_or_path: list of (label, text), eg: [(label, text), (label, text) ...]
    @param header: read_csv header
    @param names: read_csv names
    @param delimiter: read_csv sep
    @param labels_sep: multi label split
    @param is_train: is train data
    @return: X, y, data_df
    """
    if isinstance(data_list_or_path, list):
        data_df = pd.DataFrame(data_list_or_path, columns=names)
    elif isinstance(data_list_or_path, str) and os.path.exists(data_list_or_path):
        data_df = pd.read_csv(
            data_list_or_path, header=header, delimiter=delimiter, names=names
        )
    elif isinstance(data_list_or_path, pd.DataFrame):
        data_df = data_list_or_path
    else:
        raise TypeError("should be list or file path, eg: [(label, text), ... ]")
    X, y = data_df["text"], data_df["labels"]
    labels = set()
    if y.size:
        for label in y.tolist():
            if isinstance(label, str):
                labels.update(label.split(labels_sep))
            elif isinstance(label, list):
                labels.update(range(len(label)))
            else:
                labels.add(label)
        num_classes = len(labels)
        labels = sorted(list(labels))
        logger.debug(f"loaded data list, X size: {len(X)}, y size: {len(y)}")
        if is_train:
            logger.debug("num_classes: %d, labels: %s" % (num_classes, labels))
    assert len(X) == len(y)

    return X, y, data_df


class ClassifierABC:
    """
    Abstract class for classifier
    """

    def train(self, data_list_or_path, model_dir: str, **kwargs):
        raise NotImplementedError("train method not implemented.")

    def predict(self, sentences: list):
        raise NotImplementedError("predict method not implemented.")

    def evaluate_model(self, **kwargs):
        raise NotImplementedError("evaluate_model method not implemented.")

    def evaluate(self, **kwargs):
        raise NotImplementedError("evaluate method not implemented.")

    def load_model(self):
        raise NotImplementedError("load method not implemented.")

    def save_model(self):
        raise NotImplementedError("save method not implemented.")



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 1:45 PM
# @File: fasttext_model
# @Email: mlshenkai@163.com
import torch
import torch.nn as nn
import torch.nn.functional as F


class FastTextModel(nn.Module):
    """Bag of Tricks for Efficient Text Classification"""

    def __init__(
        self,
        vocab_size,
        num_classes,
        embed_size=200,
        n_gram_vocab=250499,
        hidden_size=256,
        dropout_rate=0.5,
    ):
        super().__init__()
        self.embedding = nn.Embedding(
            vocab_size, embed_size, padding_idx=vocab_size - 1
        )
        self.embedding_ngram2 = nn.Embedding(n_gram_vocab, embed_size)
        self.embedding_ngram3 = nn.Embedding(n_gram_vocab, embed_size)
        self.dropout = nn.Dropout(dropout_rate)
        self.fc1 = nn.Linear(embed_size * 3, hidden_size)
        self.fc2 = nn.Linear(hidden_size, num_classes)

    def forward(self, x):
        out_word = self.embedding(x[0])
        out_bigram = self.embedding_ngram2(x[2])
        out_trigram = self.embedding_ngram3(x[3])
        out = torch.cat((out_word, out_bigram, out_trigram), -1)

        out = out.mean(dim=1)
        out = self.dropout(out)
        out = self.fc1(out)
        out = F.relu(out)
        out = self.fc2(out)
        return out



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 1:57 PM
# @File: textcnn_classifier
# @Email: mlshenkai@163.com
import argparse
import json
import os
import sys
import time

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from loguru import logger
from sklearn import metrics
from sklearn.model_selection import train_test_split
from EasyAI.TextClassifier.base_classifier import ClassifierABC, load_data
from EasyAI.TextClassifier.data_helper import set_seed, build_vocab, load_vocab
from EasyAI.common.time_utils import get_time_spend
from .data import build_dataset, build_iterator
from .models import TextCNNModel


pwd_path = os.path.abspath(os.path.dirname(__file__))
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class TextCNNClassifier(ClassifierABC):
    def __init__(
        self,
        model_dir,
        filter_sizes=(2, 3, 4),
        num_filters=256,
        dropout_rate=0.5,
        batch_size=64,
        max_seq_length=128,
        embed_size=200,
        max_vocab_size=10000,
        unk_token="[UNK]",
        pad_token="[PAD]",
        tokenizer=None,
    ):
        """
        Init the TextCNNClassifier
        @param model_dir: 模型保存路径
        @param filter_sizes: 卷积核尺寸
        @param num_filters: 卷积核数量(channels数)
        @param dropout_rate:
        @param batch_size:
        @param max_seq_length:
        @param embed_size:
        @param max_vocab_size:
        @param unk_token:
        @param pad_token:
        @param tokenizer: 切词器，默认为字粒度切分
        """
        self.model_dir = model_dir
        self.is_trained = False
        self.model = None
        logger.debug(f"device: {device}")
        self.filter_sizes = filter_sizes
        self.num_filters = num_filters
        self.dropout_rate = dropout_rate
        self.batch_size = batch_size
        self.max_seq_length = max_seq_length
        self.embed_size = embed_size
        self.max_vocab_size = max_vocab_size
        self.unk_token = unk_token
        self.pad_token = pad_token
        self.tokenizer = (
            tokenizer if tokenizer else lambda x: [y for y in x]
        )  # char-level

    def __str__(self):
        return f"TextCNNClassifier instance ({self.model})"

    def train(
        self,
        data_list_or_path,
        header=None,
        names=("labels", "text"),
        delimiter="\t",
        test_size=0.1,
        num_epochs=20,
        learning_rate=1e-3,
        require_improvement=1000,
        evaluate_during_training_steps=100,
    ):
        """
        Train model with data_list_or_path and save model to model_dir
        @param data_list_or_path:
        @param model_dir:
        @param header:
        @param names:
        @param delimiter:
        @param test_size:
        @param num_epochs: epoch数
        @param learning_rate: 学习率
        @param require_improvement: 若超过1000batch效果还没提升，则提前结束训练
        @param evaluate_during_training_steps: 每隔多少step评估一次模型
        @return:
        """
        logger.debug("train model...")
        SEED = 1
        set_seed(SEED)
        # load data
        X, y, data_df = load_data(
            data_list_or_path,
            header=header,
            names=names,
            delimiter=delimiter,
            is_train=True,
        )
        model_dir = self.model_dir
        if model_dir:
            os.makedirs(model_dir, exist_ok=True)
        word_vocab_path = os.path.join(model_dir, "word_vocab.json")
        label_vocab_path = os.path.join(model_dir, "label_vocab.json")
        save_model_path = os.path.join(model_dir, "model.pth")

        dataset, self.word_id_map, self.label_id_map = build_dataset(
            self.tokenizer,
            X,
            y,
            word_vocab_path,
            label_vocab_path,
            max_vocab_size=self.max_vocab_size,
            max_seq_length=self.max_seq_length,
            unk_token=self.unk_token,
            pad_token=self.pad_token,
        )
        train_data, dev_data = train_test_split(
            dataset, test_size=test_size, random_state=SEED
        )
        logger.debug(
            f"train_data size: {len(train_data)}, dev_data size: {len(dev_data)}"
        )
        logger.debug(
            f"train_data sample:\n{train_data[:3]}\ndev_data sample:\n{dev_data[:3]}"
        )
        train_iter = build_iterator(train_data, device, self.batch_size)
        dev_iter = build_iterator(dev_data, device, self.batch_size)
        # create model
        vocab_size = len(self.word_id_map)
        num_classes = len(self.label_id_map)
        logger.debug(f"vocab_size:{vocab_size}", "num_classes:", num_classes)
        self.model = TextCNNModel(
            vocab_size,
            num_classes,
            embed_size=self.embed_size,
            filter_sizes=self.filter_sizes,
            num_filters=self.num_filters,
            dropout_rate=self.dropout_rate,
        )
        self.model.to(device)
        # init_network(self.model)
        logger.info(self.model.parameters)
        # train model
        history = self.train_model_from_data_iterator(
            save_model_path,
            train_iter,
            dev_iter,
            num_epochs,
            learning_rate,
            require_improvement,
            evaluate_during_training_steps,
        )
        self.is_trained = True
        logger.debug("train model done")
        return history

    def train_model_from_data_iterator(
        self,
        save_model_path,
        train_iter,
        dev_iter,
        num_epochs=10,
        learning_rate=1e-3,
        require_improvement=1000,
        evaluate_during_training_steps=100,
    ):
        history = []
        # train
        start_time = time.time()
        optimizer = torch.optim.Adam(self.model.parameters(), lr=learning_rate)

        total_batch = 0  # 记录进行到多少batch
        dev_best_loss = 1e10
        last_improve = 0  # 记录上次验证集loss下降的batch数
        flag = False  # 记录是否很久没有效果提升
        for epoch in range(num_epochs):
            logger.debug("Epoch [{}/{}]".format(epoch + 1, num_epochs))
            for i, (trains, labels) in enumerate(train_iter):
                self.model.train()
                outputs = self.model(trains)
                loss = F.cross_entropy(outputs, labels)
                # compute gradient and do SGD step
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                if total_batch % evaluate_during_training_steps == 0:
                    # 输出在训练集和验证集上的效果
                    y_true = labels.cpu()
                    y_pred = torch.max(outputs, 1)[1].cpu()
                    train_acc = metrics.accuracy_score(y_true, y_pred)
                    if dev_iter is not None:
                        dev_acc, dev_loss = self.evaluate(dev_iter)
                        if dev_loss < dev_best_loss:
                            dev_best_loss = dev_loss
                            torch.save(self.model.state_dict(), save_model_path)
                            logger.debug(f"Saved model: {save_model_path}")
                            improve = "*"
                            last_improve = total_batch
                        else:
                            improve = ""
                        time_dif = get_time_spend(start_time)
                        msg = "Iter:{0:>6},Train Loss:{1:>5.2},Train Acc:{2:>6.2%},Val Loss:{3:>5.2},Val Acc:{4:>6.2%},Time:{5} {6}".format(
                            total_batch,
                            loss.item(),
                            train_acc,
                            dev_loss,
                            dev_acc,
                            time_dif,
                            improve,
                        )
                    else:
                        time_dif = get_time_spend(start_time)
                        msg = "Iter:{0:>6},Train Loss:{1:>5.2},Train Acc:{2:>6.2%},Time:{3}".format(
                            total_batch, loss.item(), train_acc, time_dif
                        )
                    logger.debug(msg)
                    history.append(msg)
                    self.model.train()
                total_batch += 1
                if total_batch - last_improve > require_improvement:
                    # 验证集loss超过1000batch没下降，结束训练
                    logger.debug("No optimization for a long time, auto-stopping...")
                    flag = True
                    break
            if flag:
                break
        return history

    def predict(self, sentences: list):
        """
        Predict labels and label probability for sentences.
        @param sentences: list, input text list, eg: [text1, text2, ...]
        @return: predict_label, predict_prob
        """
        if not self.is_trained:
            raise ValueError("model not trained.")
        self.model.eval()

        def load_dataset(X, max_seq_length=128):
            contents = []
            for content in X:
                words_line = []
                token = self.tokenizer(content)
                seq_len = len(token)
                if max_seq_length:
                    if len(token) < max_seq_length:
                        token.extend([self.pad_token] * (max_seq_length - len(token)))
                    else:
                        token = token[:max_seq_length]
                        seq_len = max_seq_length
                # word to id
                for word in token:
                    words_line.append(
                        self.word_id_map.get(word, self.word_id_map.get(self.unk_token))
                    )
                contents.append((words_line, 0, seq_len))
            return contents

        data = load_dataset(sentences, self.max_seq_length)
        data_iter = build_iterator(data, device, self.batch_size)
        # predict prob
        predict_all = np.array([], dtype=int)
        proba_all = np.array([], dtype=float)
        with torch.no_grad():
            for texts, _ in data_iter:
                outputs = self.model(texts)
                logit = F.softmax(outputs, dim=1).detach().cpu().numpy()
                pred = np.argmax(logit, axis=1)
                proba = np.max(logit, axis=1)

                predict_all = np.append(predict_all, pred)
                proba_all = np.append(proba_all, proba)
        id_label_map = {v: k for k, v in self.label_id_map.items()}
        predict_labels = [id_label_map.get(i) for i in predict_all]
        predict_probs = proba_all.tolist()
        return predict_labels, predict_probs

    def evaluate_model(
        self, data_list_or_path, header=None, names=("labels", "text"), delimiter="\t"
    ):
        X_test, y_test, df = load_data(
            data_list_or_path, header=header, names=names, delimiter=delimiter
        )
        self.load_model()
        data, word_id_map, label_id_map = build_dataset(
            self.tokenizer,
            X_test,
            y_test,
            self.word_vocab_path,
            self.label_vocab_path,
            max_vocab_size=self.max_vocab_size,
            max_seq_length=self.max_seq_length,
            unk_token=self.unk_token,
            pad_token=self.pad_token,
        )
        data_iter = build_iterator(data, device, self.batch_size)
        return self.evaluate(data_iter)[0]

    def evaluate(self, data_iter):
        """
        Evaluate model.
        @param data_iter:
        @return: accuracy score, loss
        """
        if not self.model:
            raise ValueError("model not trained.")
        self.model.eval()
        loss_total = 0.0
        predict_all = np.array([], dtype=int)
        labels_all = np.array([], dtype=int)
        with torch.no_grad():
            for texts, labels in data_iter:
                outputs = self.model(texts)
                loss = F.cross_entropy(outputs, labels)
                loss_total += loss
                labels = labels.cpu().numpy()
                predic = torch.max(outputs, 1)[1].cpu().numpy()
                labels_all = np.append(labels_all, labels)
                predict_all = np.append(predict_all, predic)
            logger.debug(f"evaluate, last batch, y_true: {labels}, y_pred: {predic}")
        acc = metrics.accuracy_score(labels_all, predict_all)
        return acc, loss_total / len(data_iter)

    def load_model(self):
        """
        Load model from model_dir
        @return:
        """
        model_path = os.path.join(self.model_dir, "model.pth")
        if os.path.exists(model_path):
            self.word_vocab_path = os.path.join(self.model_dir, "word_vocab.json")
            self.label_vocab_path = os.path.join(self.model_dir, "label_vocab.json")
            self.word_id_map = load_vocab(self.word_vocab_path)
            self.label_id_map = load_vocab(self.label_vocab_path)
            vocab_size = len(self.word_id_map)
            num_classes = len(self.label_id_map)
            self.model = TextCNNModel(
                vocab_size,
                num_classes,
                embed_size=self.embed_size,
                filter_sizes=self.filter_sizes,
                num_filters=self.num_filters,
                dropout_rate=self.dropout_rate,
            )
            self.model.load_state_dict(torch.load(model_path, map_location=device))
            self.model.to(device)
            self.is_trained = True
        else:
            logger.error(f"{model_path} not exists.")
            self.is_trained = False
        return self.is_trained



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 1:42 PM
# @File: build_dataset
# @Email: mlshenkai@163.com
from EasyAI.TextClassifier.data_helper import load_vocab, build_vocab
from loguru import logger
import os
import json
import torch


def build_dataset(
    tokenizer,
    X,
    y,
    word_vocab_path,
    label_vocab_path,
    max_vocab_size=10000,
    max_seq_length=128,
    unk_token="[UNK]",
    pad_token="[PAD]",
    n_gram_vocab=250499,
    enable_ngram=True,
):
    if os.path.exists(word_vocab_path):
        word_id_map = load_vocab(word_vocab_path)
    else:
        word_id_map = build_vocab(
            X,
            tokenizer=tokenizer,
            max_size=max_vocab_size,
            min_freq=1,
            unk_token=unk_token,
            pad_token=pad_token,
        )
        json.dump(
            word_id_map,
            open(word_vocab_path, "w", encoding="utf-8"),
            ensure_ascii=False,
            indent=4,
        )
        logger.debug("save vocab_path: {}".format(word_vocab_path))
    logger.debug(
        f"word vocab size: {len(word_id_map)}, word_vocab_path: {word_vocab_path}"
    )

    if os.path.exists(label_vocab_path):
        label_id_map = load_vocab(label_vocab_path)
    else:
        id_label_map = {id: v for id, v in enumerate(set(y.tolist()))}
        label_id_map = {v: k for k, v in id_label_map.items()}
        json.dump(
            label_id_map,
            open(label_vocab_path, "w", encoding="utf-8"),
            ensure_ascii=False,
            indent=4,
        )
        logger.debug("save label_vocab_path: {}".format(label_vocab_path))
    logger.debug(
        f"label vocab size: {len(label_id_map)}, label_vocab_path: {label_vocab_path}"
    )

    def biGramHash(sequence, t, buckets):
        t1 = sequence[t - 1] if t - 1 >= 0 else 0
        return (t1 * 14918087) % buckets

    def triGramHash(sequence, t, buckets):
        t1 = sequence[t - 1] if t - 1 >= 0 else 0
        t2 = sequence[t - 2] if t - 2 >= 0 else 0
        return (t2 * 14918087 * 18408749 + t1 * 14918087) % buckets

    def load_dataset(X, y, max_seq_length=128):
        contents = []
        for content, label in zip(X, y):
            words_line = []
            token = tokenizer(content)
            seq_len = len(token)
            if max_seq_length:
                if len(token) < max_seq_length:
                    token.extend([pad_token] * (max_seq_length - len(token)))
                else:
                    token = token[:max_seq_length]
                    seq_len = max_seq_length
            # word to id
            for word in token:
                words_line.append(word_id_map.get(word, word_id_map.get(unk_token)))
            label_id = label_id_map.get(label)
            # fasttext ngram
            bigram = []
            trigram = []
            if enable_ngram:
                buckets = n_gram_vocab
                # ------ngram------
                for i in range(max_seq_length):
                    bigram.append(biGramHash(words_line, i, buckets))
                    trigram.append(triGramHash(words_line, i, buckets))
                # -----------------
            else:
                bigram = [0] * max_seq_length
                trigram = [0] * max_seq_length
            contents.append((words_line, label_id, seq_len, bigram, trigram))
        return contents

    dataset = load_dataset(X, y, max_seq_length)
    return dataset, word_id_map, label_id_map


class DatasetIterater:
    def __init__(self, dataset, device, batch_size=32):
        self.batch_size = batch_size
        self.dataset = dataset
        self.n_batches = len(dataset) // batch_size if len(dataset) > batch_size else 1
        self.residue = False  # 记录batch数量是否为整数
        if len(dataset) % self.n_batches != 0:
            self.residue = True
        self.index = 0
        self.device = device

    def _to_tensor(self, datas):
        x = torch.LongTensor([_[0] for _ in datas]).to(self.device)
        y = torch.LongTensor([_[1] for _ in datas]).to(self.device)
        bigram = torch.LongTensor([_[3] for _ in datas]).to(self.device)
        trigram = torch.LongTensor([_[4] for _ in datas]).to(self.device)

        # pad_token前的长度(超过max_seq_length的设为max_seq_length)
        seq_len = torch.LongTensor([_[2] for _ in datas]).to(self.device)
        return (x, seq_len, bigram, trigram), y

    def __next__(self):
        if self.residue and self.index == self.n_batches:
            batches = self.dataset[self.index * self.batch_size : len(self.dataset)]
            self.index += 1
            batches = self._to_tensor(batches)
            return batches

        elif self.index >= self.n_batches:
            self.index = 0
            raise StopIteration
        else:
            batches = self.dataset[
                self.index * self.batch_size : (self.index + 1) * self.batch_size
            ]
            self.index += 1
            batches = self._to_tensor(batches)
            return batches

    def __iter__(self):
        return self

    def __len__(self):
        if self.residue:
            return self.n_batches + 1
        else:
            return self.n_batches


def build_iterator(dataset, device, batch_size=32):
    return DatasetIterater(dataset, device, batch_size)



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 2:05 PM
# @File: textrnn_model
# @Email: mlshenkai@163.com
import torch
import torch.nn as nn
import torch.nn.functional as F


class TextRNNAttModel(nn.Module):
    """Attention-Based Bidirectional Long Short-Term Memory Networks for Relation Classification"""

    def __init__(
        self,
        vocab_size,
        num_classes,
        embed_size=200,
        hidden_size=128,
        num_layers=2,
        dropout_rate=0.5,
    ):
        """
        TextRNN-Att model
        @param vocab_size:
        @param num_classes:
        @param embed_size: 字向量维度
        @param hidden_size: lstm隐藏层
        @param num_layers: lstm层数
        @param dropout_rate:
        """
        super().__init__()
        self.embedding = nn.Embedding(
            vocab_size, embed_size, padding_idx=vocab_size - 1
        )
        self.lstm = nn.LSTM(
            embed_size,
            hidden_size,
            num_layers,
            bidirectional=True,
            batch_first=True,
            dropout=dropout_rate,
        )
        self.tanh1 = nn.Tanh()
        # self.u = nn.Parameter(torch.Tensor(config.hidden_size * 2, config.hidden_size * 2))
        self.w = nn.Parameter(torch.zeros(hidden_size * 2))
        self.tanh2 = nn.Tanh()
        self.fc1 = nn.Linear(hidden_size * 2, int(hidden_size / 2))
        self.fc2 = nn.Linear(int(hidden_size / 2), num_classes)

    def forward(self, x):
        emb = self.embedding(x[0])  # [batch_size, seq_len, embeding]=[128, 32, 300]
        H, _ = self.lstm(
            emb
        )  # [batch_size, seq_len, hidden_size * num_direction]=[128, 32, 256]

        M = self.tanh1(H)  # [128, 32, 256]
        # M = torch.tanh(torch.matmul(H, self.u))
        alpha = F.softmax(torch.matmul(M, self.w), dim=1).unsqueeze(-1)  # [128, 32, 1]
        out = H * alpha  # [128, 32, 256]
        out = torch.sum(out, 1)  # [128, 256]
        out = F.relu(out)
        out = self.fc1(out)
        out = self.fc2(out)  # [128, 64]
        return out



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 1:53 PM
# @File: textcnn_model
# @Email: mlshenkai@163.com
import torch
import torch.nn as nn
import torch.nn.functional as F


class TextCNNModel(nn.Module):
    """Convolutional Neural Networks for Sentence Classification"""

    def __init__(
        self,
        vocab_size,
        num_classes,
        embed_size=200,
        filter_sizes=(2, 3, 4),
        num_filters=256,
        dropout_rate=0.5,
    ):
        """
        Init the TextCNNModel
        @param vocab_size:
        @param num_classes:
        @param embed_size:
        @param filter_sizes: 卷积核尺寸
        @param num_filters: 卷积核数量(channels数)
        @param dropout_rate:
        """
        super().__init__()
        self.embedding = nn.Embedding(
            vocab_size, embed_size, padding_idx=vocab_size - 1
        )
        self.convs = nn.ModuleList(
            [nn.Conv2d(1, num_filters, (k, embed_size)) for k in filter_sizes]
        )
        self.dropout = nn.Dropout(dropout_rate)
        self.fc = nn.Linear(num_filters * len(filter_sizes), num_classes)

    def conv_and_pool(self, x, conv):
        x = F.relu(conv(x)).squeeze(3)
        x = F.max_pool1d(x, x.size(2)).squeeze(2)
        return x

    def forward(self, x):
        out = self.embedding(x[0])
        out = out.unsqueeze(1)
        out = torch.cat([self.conv_and_pool(out, conv) for conv in self.convs], 1)
        out = self.dropout(out)
        out = self.fc(out)
        return out



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 2:18 PM
# @File: text_cluster
# @Email: mlshenkai@163.com
import os
from codecs import open
import pickle
from sklearn.cluster import MiniBatchKMeans
from sklearn.feature_extraction.text import TfidfVectorizer
from loguru import logger
from EasyAI.TextClassifier.tokenizer import Tokenizer

pwd_path = os.path.abspath(os.path.dirname(__file__))
default_stopwords_path = os.path.join(pwd_path, "stopwords.txt")


class TextCluster(object):
    def __init__(
        self,
        model_dir,
        model=None,
        tokenizer=None,
        feature=None,
        stopwords_path=default_stopwords_path,
        n_clusters=3,
        n_init=10,
        ngram_range=(1, 2),
        **kwargs
    ):
        self.model_dir = model_dir
        self.model = (
            model if model else MiniBatchKMeans(n_clusters=n_clusters, n_init=n_init)
        )
        self.tokenizer = tokenizer if tokenizer else Tokenizer()
        self.feature = (
            feature if feature else TfidfVectorizer(ngram_range=ngram_range, **kwargs)
        )
        self.stopwords = (
            set(self.load_list(stopwords_path))
            if stopwords_path and os.path.exists(stopwords_path)
            else set()
        )
        self.is_trained = False

    def __str__(self):
        return "TextCluster instance ({}, {}, {})".format(
            self.model, self.tokenizer, self.feature
        )

    @staticmethod
    def load_file_data(file_path, sep="\t", use_col=1):
        """
        Load text file, format(txt): text
        :param file_path: str
        :param sep: \t
        :param use_col: int or None
        :return: list, text list
        """
        contents = []
        if not os.path.exists(file_path):
            raise ValueError("file not found. path: {}".format(file_path))
        with open(file_path, "r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if use_col:
                    contents.append(line.split(sep)[use_col])
                else:
                    contents.append(line)
        logger.info(
            "load file done. path: {}, size: {}".format(file_path, len(contents))
        )
        return contents

    @staticmethod
    def load_list(path):
        """
        加载停用词
        :param path:
        :return: list
        """
        return [word for word in open(path, "r", encoding="utf-8").read().split()]

    @staticmethod
    def load_pkl(pkl_path):
        """
        加载词典文件
        :param pkl_path:
        :return:
        """
        with open(pkl_path, "rb") as f:
            result = pickle.load(f)
        return result

    @staticmethod
    def save_pkl(vocab, pkl_path, overwrite=True):
        """
        存储文件
        :param pkl_path:
        :param overwrite:
        :return:
        """
        if pkl_path and os.path.exists(pkl_path) and not overwrite:
            return
        if pkl_path:
            with open(pkl_path, "wb") as f:
                pickle.dump(vocab, f, protocol=pickle.HIGHEST_PROTOCOL)
                # pickle.dump(vocab, f, protocol=2)  # 兼容python2和python3
            print("save %s ok." % pkl_path)

    @staticmethod
    def show_plt(feature_matrix, labels, image_file="cluster.png"):
        """
        Show cluster plt
        :param feature_matrix:
        :param labels:
        :param image_file:
        :return:
        """
        from sklearn.decomposition import TruncatedSVD
        import matplotlib.pyplot as plt

        svd = TruncatedSVD()
        plot_columns = svd.fit_transform(feature_matrix)
        plt.scatter(x=plot_columns[:, 0], y=plot_columns[:, 1], c=labels)
        if image_file:
            plt.savefig(image_file)
        plt.show()

    def tokenize_sentences(self, sentences):
        """
        Encoding input text
        :param sentences: list of text, eg: [text1, text2, ...]
        :return: X_tokens
        """
        X_tokens = [
            " ".join(
                [w for w in self.tokenizer.tokenize(line) if w not in self.stopwords]
            )
            for line in sentences
        ]
        return X_tokens

    def train(self, sentences):
        """
        Train model and save model
        :param sentences: list of text, eg: [text1, text2, ...]
        :return: model
        """
        logger.debug("train model")
        X_tokens = self.tokenize_sentences(sentences)
        logger.debug("data tokens top 1: {}".format(X_tokens[:1]))
        feature = self.feature.fit_transform(X_tokens)
        # fit cluster
        self.model.fit(feature)
        labels = self.model.labels_
        logger.debug("cluster labels:{}".format(labels))
        model_dir = self.model_dir
        if model_dir:
            os.makedirs(model_dir, exist_ok=True)
        feature_path = os.path.join(model_dir, "cluster_feature.pkl")
        self.save_pkl(self.feature, feature_path)
        model_path = os.path.join(model_dir, "cluster_model.pkl")
        self.save_pkl(self.model, model_path)
        logger.info(
            "save done. feature path: {}, model path: {}".format(
                feature_path, model_path
            )
        )

        self.is_trained = True
        return feature, labels

    def predict(self, X):
        """
        Predict label
        :param X: list, input text list, eg: [text1, text2, ...]
        :return: list, label name
        """
        if not self.is_trained:
            raise ValueError("model is None, run train first.")
        # tokenize text
        X_tokens = self.tokenize_sentences(X)
        # transform
        feat = self.feature.transform(X_tokens)
        return self.model.predict(feat)

    def show_clusters(self, feature_matrix, labels, image_file="cluster.png"):
        """
        Show cluster plt image
        :param feature_matrix:
        :param labels:
        :param image_file:
        :return:
        """
        if not self.is_trained:
            raise ValueError("model is None, run train first.")
        self.show_plt(feature_matrix, labels, image_file)

    def load_model(self):
        """
        Load model from model_dir
        :param model_dir: path
        :return: None
        """
        model_path = os.path.join(self.model_dir, "cluster_model.pkl")
        if not os.path.exists(model_path):
            raise ValueError("model is not found. please train and save model first.")
        self.model = self.load_pkl(model_path)
        feature_path = os.path.join(self.model_dir, "cluster_feature.pkl")
        self.feature = self.load_pkl(feature_path)
        self.is_trained = True
        logger.info("model loaded {}".format(self.model_dir))
        return self.is_trained



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 11:57 AM
# @File: time_utils
# @Email: mlshenkai@163.com

import time
from datetime import timedelta


def get_time_spend(start_time):
    """获取已使用时间"""
    end_time = time.time()
    time_dif = end_time - start_time
    return timedelta(seconds=int(round(time_dif)))


def init_network(model, method="xavier", exclude="embedding"):
    """权重初始化，默认xavier"""
    import torch.nn as nn

    for name, w in model.named_parameters():
        if exclude not in name:
            if "weight" in name:
                if method == "xavier":
                    nn.init.xavier_normal_(w)
                elif method == "kaiming":
                    nn.init.kaiming_normal_(w)
                else:
                    nn.init.normal_(w)
            if "bias" in name:
                nn.init.constant_(w, 0)



# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/24 4:31 PM
# @File: setup
# @Email: mlshenkai@163.com

from distutils.core import setup
from setuptools import find_packages
with open("README.md", "r") as f:
    long_description = f.read()


setup(
    name="EasyAIScaffolding",
    version="1.0.0",
    description="A PyTorch-based AI development scaffolding",
    long_description=long_description,
    author="watcher shen",
    author_email="watcher.shen@gmail.com",
    url="https://github.com/mlshenkai/EasyAIScaffolding",
    license="MIT License",
    packages=find_packages(),
    platforms=["all"],
    classifiers=[]
)


# -*- coding: utf-8 -*-
# @Author: watcher
# @Created Time: 2023/3/23 4:17 PM
# @File: data_helper
# @Email: mlshenkai@163.com

import json
import random

import numpy as np
import torch
from loguru import logger
from tqdm import tqdm


def set_seed(seed):
    """
    Set seed for random number generators.
    """
    logger.info(f"Set seed for random, numpy and torch: {seed}")
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)


def build_vocab(contents, tokenizer, max_size, min_freq, unk_token, pad_token):
    vocab_dic = {}
    for line in tqdm(contents):
        line = line.strip()
        if not line:
            continue
        content = line.split("\t")[0]
        for word in tokenizer(content):
            vocab_dic[word] = vocab_dic.get(word, 0) + 1
    vocab_list = sorted(
        [_ for _ in vocab_dic.items() if _[1] >= min_freq],
        key=lambda x: x[1],
        reverse=True,
    )[:max_size]
    vocab_dic = {word_count[0]: idx for idx, word_count in enumerate(vocab_list)}
    vocab_dic.update({unk_token: len(vocab_dic), pad_token: len(vocab_dic) + 1})
    return vocab_dic


def load_vocab(vocab_path):
    """
    :param vocab_path:
    :return:
    """
    with open(vocab_path, "r", encoding="utf-8") as fr:
        vocab = json.load(fr)
    return vocab
